import {
  RDSClient,
  DescribeDBClustersCommand
} from '@aws-sdk/client-rds';
import { BPSet, BPSetMetadata, BPSetStats } from '../../types';
import { Memorizer } from '../../Memorizer';

export class RDSClusterMultiAZEnabled implements BPSet {
  private readonly client = new RDSClient({});
  private readonly memoClient = Memorizer.memo(this.client);

  private readonly stats: BPSetStats = {
    compliantResources: [],
    nonCompliantResources: [],
    status: 'LOADED',
    errorMessage: []
  };

  public readonly getMetadata = (): BPSetMetadata => ({
    name: 'RDSClusterMultiAZEnabled',
    description: 'Ensures that RDS clusters are deployed across multiple availability zones.',
    priority: 1,
    priorityReason: 'Multi-AZ deployment improves availability and resilience of RDS clusters.',
    awsService: 'RDS',
    awsServiceCategory: 'Database',
    bestPracticeCategory: 'Availability',
    requiredParametersForFix: [
      {
        name: 'additional-azs',
        description: 'Number of additional availability zones to add for Multi-AZ configuration.',
        default: '2',
        example: '2'
      }
    ],
    isFixFunctionUsesDestructiveCommand: false,
    commandUsedInCheckFunction: [
      {
        name: 'DescribeDBClustersCommand',
        reason: 'To fetch the list of RDS clusters and their availability zone configuration.'
      }
    ],
    commandUsedInFixFunction: [],
    adviseBeforeFixFunction:
      'Enabling Multi-AZ for an existing cluster may require significant reconfiguration and potential downtime. Proceed with caution.'
  });

  public readonly getStats = () => this.stats;

  public readonly clearStats = () => {
    this.stats.compliantResources = [];
    this.stats.nonCompliantResources = [];
    this.stats.status = 'LOADED';
    this.stats.errorMessage = [];
  };

  public readonly check = async () => {
    this.stats.status = 'CHECKING';

    await this.checkImpl()
      .then(() => {
        this.stats.status = 'FINISHED';
      })
      .catch((err) => {
        this.stats.status = 'ERROR';
        this.stats.errorMessage.push({
          date: new Date(),
          message: err.message
        });
      });
  };

  private readonly checkImpl = async () => {
    const compliantResources: string[] = [];
    const nonCompliantResources: string[] = [];
    const dbClusters = await this.getDBClusters();

    for (const cluster of dbClusters) {
      if ((cluster.AvailabilityZones || []).length > 1) {
        compliantResources.push(cluster.DBClusterArn!);
      } else {
        nonCompliantResources.push(cluster.DBClusterArn!);
      }
    }

    this.stats.compliantResources = compliantResources;
    this.stats.nonCompliantResources = nonCompliantResources;
  };

  public readonly fix = async (
    nonCompliantResources: string[],
    requiredParametersForFix: { name: string; value: string }[]
  ) => {
    this.stats.status = 'CHECKING';

    await this.fixImpl(nonCompliantResources, requiredParametersForFix)
      .then(() => {
        this.stats.status = 'FINISHED';
      })
      .catch((err) => {
        this.stats.status = 'ERROR';
        this.stats.errorMessage.push({
          date: new Date(),
          message: err.message
        });
      });
  };

  private readonly fixImpl = async (
    nonCompliantResources: string[],
    requiredParametersForFix: { name: string; value: string }[]
  ) => {
    throw new Error(
      'Enabling Multi-AZ requires cluster reconfiguration. This must be performed manually.'
    );
  };

  private readonly getDBClusters = async () => {
    const response = await this.memoClient.send(new DescribeDBClustersCommand({}));
    return response.DBClusters || [];
  };
}
